"""Basic NanoEvents and NanoCollection mixins"""
import awkward1


behavior = {}


@awkward1.mixin_class(behavior)
class NanoEvents:
    """NanoEvents mixin class

    This mixin class is used as the top-level type for NanoEvents objects.
    """

    @property
    def metadata(self):
        """Arbitrary metadata"""
        return self.layout.purelist_parameter("metadata")


behavior[("__typestr__", "NanoEvents")] = "event"


@awkward1.mixin_class(behavior)
class NanoCollection:
    """A NanoEvents collection

    This mixin provides some helper methods useful for creating cross-references
    and other advanced mixin types.
    """

    def _getlistarray(self):
        """Do some digging to find the initial listarray"""

        def descend(layout, depth):
            islistarray = isinstance(
                layout,
                (
                    awkward1.layout.ListOffsetArray32,
                    awkward1.layout.ListOffsetArray64,
                ),
            )
            if islistarray and layout.content.parameter("collection_name") is not None:
                return lambda: layout

        return awkward1._util.recursively_apply(self.layout, descend)

    def _content(self):
        """Internal method to get jagged collection content

        This should only be called on the original unsliced collection array.
        Used with global indexes to resolve cross-references"""
        return self._getlistarray().content

    def _apply_global_index(self, index):
        """Internal method to take from a collection using a flat index

        This is often necessary to be able to still resolve cross-references on
        reduced arrays or single records.
        """
        if isinstance(index, int):
            out = self._content()[index]
        else:

            def flat_take(layout):
                idx = awkward1.Array(layout)
                return self._content()[idx.mask[idx >= 0]]

            def descend(layout, depth):
                if layout.purelist_depth == 1:
                    return lambda: flat_take(layout)

            (index,) = awkward1.broadcast_arrays(index)
            out = awkward1._util.recursively_apply(index.layout, descend)
        return awkward1._util.wrap(out, self.behavior, cache=self.cache)

    def _events(self):
        """Internal method to get the originally-constructed NanoEvents

        This can be called at any time from any collection, as long as
        the NanoEventsFactory instance exists."""
        return self.behavior["__events_factory__"].events()



"""NanoEvents and helpers

"""
from coffea.nanoevents.factory import NanoEventsFactory
from coffea.nanoevents.schemas import BaseSchema, NanoAODSchema, TreeMakerSchema

__all__ = [
    "NanoEventsFactory",
    "BaseSchema",
    "NanoAODSchema",
    "TreeMakerSchema",
]



from coffea import hist, processor
from coffea.analysis_objects import JaggedCandidateArray as CandArray
from coffea.util import awkward as akd
from coffea.util import numpy as np


class NanoTestProcessor(processor.ProcessorABC):
    def __init__(self, columns=[]):
        self._columns = columns
        dataset_axis = hist.Cat("dataset", "Primary dataset")
        mass_axis = hist.Bin("mass", r"$m_{\mu\mu}$ [GeV]", 30000, 0.25, 300)
        pt_axis = hist.Bin("pt", r"$p_{T}$ [GeV]", 30000, 0.25, 300)

        self._accumulator = processor.dict_accumulator(
            {
                "mass": hist.Hist("Counts", dataset_axis, mass_axis),
                "pt": hist.Hist("Counts", dataset_axis, pt_axis),
                "cutflow": processor.defaultdict_accumulator(int),
            }
        )

    @property
    def columns(self):
        return self._columns

    @property
    def accumulator(self):
        return self._accumulator

    def process(self, df):
        output = self.accumulator.identity()

        dataset = df["dataset"]

        muon = None
        if isinstance(df["Muon_pt"], akd.JaggedArray):
            muon = CandArray.candidatesfromcounts(
                counts=df["Muon_pt"].counts,
                pt=df["Muon_pt"].content,
                eta=df["Muon_eta"].content,
                phi=df["Muon_phi"].content,
                mass=df["Muon_mass"].content,
            )
        else:
            muon = CandArray.candidatesfromcounts(
                counts=df["nMuon"],
                pt=df["Muon_pt"],
                eta=df["Muon_eta"],
                phi=df["Muon_phi"],
                mass=df["Muon_mass"],
            )

        dimuon = muon.distincts()

        output["pt"].fill(dataset=dataset, pt=muon.pt.flatten())
        output["mass"].fill(dataset=dataset, mass=dimuon.mass.flatten())
        output["cutflow"]["%s_pt" % dataset] += np.sum(muon.counts)
        output["cutflow"]["%s_mass" % dataset] += np.sum(dimuon.counts)

        return output

    def postprocess(self, accumulator):
        return accumulator



"""BTag tools: CMS analysis-level b-tagging corrections and uncertainties

These classes provide computation of CMS b-tagging and mistagging
corrections and uncertainties on columnar data.
"""
from .btagscalefactor import BTagScaleFactor

__all__ = [
    "BTagScaleFactor",
]



import pandas
import numpy
from coffea.lookup_tools.dense_mapped_lookup import dense_mapped_lookup


class BTagScaleFactor:
    """A class holding one complete BTag scale factor for a given working point

    Parameters
    ----------
        filename : str
            The BTag-formatted CSV file to read (accepts .csv, .csv.gz, etc.)
            See pandas read_csv for all supported compressions.
        workingpoint : str or int
            The working point, one of LOOSE, MEDIUM, TIGHT, or RESHAPE (0-3, respectively)
        methods : str, optional
            The scale factor derivation method to use for each flavor, 'b,c,light' respectively.
            Defaults to 'comb,comb,incl'
        keep_df : bool, optional
            If set true, keep the parsed dataframe as an attribute (.df) for later inspection
    """

    LOOSE, MEDIUM, TIGHT, RESHAPE = range(4)
    FLAV_B, FLAV_C, FLAV_UDSG = range(3)
    _flavor = numpy.array([0, 4, 5, 6])
    _flavor2btvflavor = {0: FLAV_UDSG, 4: FLAV_C, 5: FLAV_B}
    _wpString = {"loose": LOOSE, "medium": MEDIUM, "tight": TIGHT, "reshape": RESHAPE}
    _expectedColumns = [
        "OperatingPoint",
        "measurementType",
        "sysType",
        "jetFlavor",
        "etaMin",
        "etaMax",
        "ptMin",
        "ptMax",
        "discrMin",
        "discrMax",
        "formula",
    ]

    @classmethod
    def readcsv(cls, filename):
        """Reads a BTag-formmated CSV file into a pandas dataframe

        This function also merges the bin min and max into a tuple representing the bin

        Parameters
        ----------
            filename : str
                The file to open

        Returns
        -------
            df : pandas.DataFrame
                A dataframe containing all info in the file
            discriminator : str
                The name of the discriminator the correction map is for
        """
        df = pandas.read_csv(filename, skipinitialspace=True)
        discriminator = df.columns[0].split(";")[0]

        def cleanup(colname):
            if ";" in colname:
                _, colname = colname.split(";")
            return colname.strip()

        df.rename(columns=cleanup, inplace=True)
        if not list(df.columns) == BTagScaleFactor._expectedColumns:
            raise RuntimeError(
                "Columns in BTag scale factor file %s as expected" % filename
            )
        for var in ["eta", "pt", "discr"]:
            df[var + "Bin"] = list(zip(df[var + "Min"], df[var + "Max"]))
            del df[var + "Min"]
            del df[var + "Max"]
        return df, discriminator

    def __init__(self, filename, workingpoint, methods="comb,comb,incl", keep_df=False):
        if workingpoint not in [0, 1, 2, 3]:
            try:
                workingpoint = BTagScaleFactor._wpString[workingpoint.lower()]
            except (KeyError, AttributeError):
                raise ValueError("Unrecognized working point")
        methods = methods.split(",")
        self.workingpoint = workingpoint
        df, self.discriminator = BTagScaleFactor.readcsv(filename)
        cut = (df["jetFlavor"] == self.FLAV_B) & (df["measurementType"] == methods[0])
        if len(methods) > 1:
            cut |= (df["jetFlavor"] == self.FLAV_C) & (
                df["measurementType"] == methods[1]
            )
        if len(methods) > 2:
            cut |= (df["jetFlavor"] == self.FLAV_UDSG) & (
                df["measurementType"] == methods[2]
            )
        cut &= df["OperatingPoint"] == workingpoint
        df = df[cut]
        mavailable = list(df["measurementType"].unique())
        if not all(m in mavailable for m in methods):
            raise ValueError(
                "Unrecognized jet correction method, available: %r" % mavailable
            )
        df = df.set_index(
            ["sysType", "jetFlavor", "etaBin", "ptBin", "discrBin"]
        ).sort_index()
        if keep_df:
            self.df = df
        self._corrections = {}
        for syst in list(df.index.levels[0]):
            corr = df.loc[syst]
            allbins = list(corr.index)
            # NOTE: here we force the class to assume abs(eta) based on lack of examples where signed eta is used
            edges_eta = numpy.array(
                sorted(set(abs(x) for tup in corr.index.levels[1] for x in tup))
            )
            edges_pt = numpy.array(
                sorted(set(x for tup in corr.index.levels[2] for x in tup))
            )
            edges_discr = numpy.array(
                sorted(set(x for tup in corr.index.levels[3] for x in tup))
            )
            bin_low_edges = numpy.meshgrid(
                self._flavor[:-1],
                edges_eta[:-1],
                edges_pt[:-1],
                edges_discr[:-1],
                indexing="ij",
            )
            mapping = numpy.full(bin_low_edges[0].shape, -1)

            def findbin(flavor, eta, pt, discr):
                btvflavor = self._flavor2btvflavor[flavor]
                for i, (fbin, ebin, pbin, dbin) in enumerate(allbins):
                    if (
                        btvflavor == fbin
                        and ebin[0] <= eta < ebin[1]
                        and pbin[0] <= pt < pbin[1]
                        and dbin[0] <= discr < dbin[1]
                    ):
                        return i
                return -1

            for idx, _ in numpy.ndenumerate(mapping):
                flavor, eta, pt, discr = (x[idx] for x in bin_low_edges)
                mapping[idx] = findbin(flavor, eta, pt, discr)

            if self.workingpoint == BTagScaleFactor.RESHAPE:
                self._corrections[syst] = dense_mapped_lookup(
                    (self._flavor, edges_eta, edges_pt, edges_discr),
                    mapping,
                    numpy.array(corr["formula"]),
                    3,
                )
            else:
                self._corrections[syst] = dense_mapped_lookup(
                    (self._flavor, edges_eta, edges_pt),
                    mapping[..., 0],
                    numpy.array(corr["formula"]),
                    2,
                )

    def eval(self, systematic, flavor, abseta, pt, discr=None, ignore_missing=False):
        """Evaluate this scale factor as a function of jet properties

        Parameters
        ----------
            systematic : str
                Which systematic to evaluate. Nominal correction is 'central', the options depend
                on the scale factor and method
            flavor : numpy.ndarray or awkward.Array
                The generated jet hadron flavor, following the enumeration:
                0: uds quark or gluon, 4: charm quark, 5: bottom quark
            abseta : numpy.ndarray or awkward.Array
                The absolute value of the jet pseudorapitiy
            pt : numpy.ndarray or awkward.Array
                The jet transverse momentum
            discr : numpy.ndarray or awkward.Array, optional
                The jet tagging discriminant value (default None), optional for all scale factors except
                the reshaping scale factor
            ignore_missing : bool, optional
                If set true, any values that have no correction will return 1. instead of throwing
                an exception. Out-of-bounds values are always clipped to the nearest bin.

        Returns
        -------
            out : numpy.ndarray or awkward.Array
                An array with shape matching ``pt``, containing the per-jet scale factor
        """
        if systematic not in self._corrections:
            raise ValueError("Unrecognized systematic: %s" % systematic)
        if self.workingpoint == BTagScaleFactor.RESHAPE:
            if discr is None:
                raise ValueError("RESHAPE scale factor requires a discriminant array")
            return self._corrections[systematic](
                flavor, abseta, pt, discr, ignore_missing=ignore_missing
            )
        else:
            return self._corrections[systematic](
                flavor, abseta, pt, ignore_missing=ignore_missing
            )

    def __call__(
        self, systematic, flavor, abseta, pt, discr=None, ignore_missing=False
    ):
        return self.eval(systematic, flavor, abseta, pt, discr, ignore_missing)



from cachetools import LRUCache
from collections.abc import Mapping
import uproot4
import numpy
from coffea.nanoevents import transforms
from coffea.nanoevents.util import key_to_tuple, tuple_to_key


class TrivialOpener:
    def __init__(self, uuid_pfnmap, uproot_options={}):
        self._uuid_pfnmap = uuid_pfnmap
        self._uproot_options = uproot_options

    def open_uuid(self, uuid):
        pfn = self._uuid_pfnmap[uuid]
        rootdir = uproot4.open(pfn, **self._uproot_options)
        if str(rootdir.file.uuid) != uuid:
            raise RuntimeError(
                f"UUID of file {pfn} does not match expected value ({uuid})"
            )
        return rootdir


class UprootSourceMapping(Mapping):
    _debug = False

    def __init__(self, fileopener, cache=None, access_log=None):
        self._fileopener = fileopener
        self._cache = cache
        self._access_log = access_log
        self.setup()

    def setup(self):
        if self._cache is None:
            self._cache = LRUCache(1)

    def __getstate__(self):
        return {
            "fileopener": self._fileopener,
        }

    def __setstate__(self, state):
        self._fileopener = state["fileopener"]
        self._cache = None
        self.setup()

    def _tree(self, uuid, treepath):
        key = "UprootSourceMapping:" + tuple_to_key((uuid, treepath))
        try:
            return self._cache[key]
        except KeyError:
            pass
        tree = self._fileopener.open_uuid(uuid)[treepath]
        self._cache[key] = tree
        return tree

    def preload_tree(self, uuid, treepath, tree):
        """To save a double-open when using NanoEventsFactory.from_file"""
        key = "UprootSourceMapping:" + tuple_to_key((uuid, treepath))
        self._cache[key] = tree

    @classmethod
    def interpret_key(cls, key):
        uuid, treepath, entryrange, form_key, *layoutattr = key_to_tuple(key)
        start, stop = (int(x) for x in entryrange.split("-"))
        nodes = form_key.split(",")
        if len(layoutattr) == 1:
            nodes.append("!" + layoutattr[0])
        elif len(layoutattr) > 1:
            raise RuntimeError(f"Malformed key: {key}")
        return uuid, treepath, start, stop, nodes

    def __getitem__(self, key):
        uuid, treepath, start, stop, nodes = UprootSourceMapping.interpret_key(key)
        if UprootSourceMapping._debug:
            print("Gettting:", uuid, treepath, start, stop, nodes)
        stack = []
        skip = False
        for node in nodes:
            if skip:
                skip = False
                continue
            elif node == "!skip":
                skip = True
                continue
            elif node == "!load":
                branch = stack.pop()
                if self._access_log is not None:
                    self._access_log.append(branch)
                branch = self._tree(uuid, treepath)[branch]
                # make sure uproot is single-core since our calling context might not be
                stack.append(
                    branch.array(
                        entry_start=start,
                        entry_stop=stop,
                        decompression_executor=uproot4.source.futures.TrivialExecutor(),
                        interpretation_executor=uproot4.source.futures.TrivialExecutor(),
                    )
                )
            elif node.startswith("!"):
                tname = node[1:]
                if not hasattr(transforms, tname):
                    raise RuntimeError(
                        f"Syntax error in form_key: no transform named {tname}"
                    )
                getattr(transforms, tname)(stack)
            else:
                stack.append(node)
        if len(stack) != 1:
            raise RuntimeError(f"Syntax error in form key {nodes}")
        out = stack.pop()
        try:
            out = numpy.array(out)
        except ValueError:
            if UprootSourceMapping._debug:
                print(out)
            raise RuntimeError(
                f"Left with non-bare array after evaluating form key {nodes}"
            )
        return out

    def __len__(self):
        raise NotImplementedError

    def __iter__(self):
        raise NotImplementedError


class CachedMapping(Mapping):
    """A cache-wrapped mapping

    Reads will call into ``cache`` first, and if no key exists,
    the read will fall back to ``base``, saving the reult into ``cache``.
    """

    def __init__(self, cache, base):
        self.cache = cache
        self.base = base
        self.stats = {"hit": 0, "miss": 0}

    def __getitem__(self, key):
        try:
            value = self.cache[key]
            self.stats["hit"] += 1
            return value
        except KeyError:
            value = self.base[key]
            self.cache[key] = value
            self.stats["miss"] += 1
            return value

    def __iter__(self):
        return iter(self.base)

    def __len__(self):
        return len(self.base)



from ..util import numpy as np
import uproot
import re

cycle = re.compile(br";\d+")


def killcycle(s, cycle):
    return cycle.sub(b"", s)


histTypes = [b"TH1D", b"TH1F", b"TH2D", b"TH2F", b"TH3D", b"TH3F"]
graphTypes = [b"TGraphAsymmErrors", b"TGraph2D"]


def convert_histo_root_file(file):
    converted_file = {}
    fin = uproot.open(file.strip())
    for path, item in fin.iteritems(recursive=True):
        nicepath = killcycle(path, cycle).decode()
        rootclass = item._classname
        if rootclass in histTypes:
            converted_file[(nicepath, "dense_lookup")] = item.values, item.edges
            if hasattr(item, "variances"):
                converted_file[(nicepath + "_error", "dense_lookup")] = (
                    np.sqrt(item.variances),
                    item.edges,
                )
        elif rootclass in graphTypes:
            # TODO: convert TGraph into interpolated lookup
            continue
        elif hasattr(item, "_fEX") and hasattr(item, "_fEY"):
            # TODO what is this?
            tempArrX = item._fEX
            tempArrY = item._fEY
            converted_file[(nicepath, "dense_lookup")] = [tempArrX, tempArrY]

    return converted_file



import numpy
import awkward
import numba
from .common import LorentzVector, Candidate
from ..util import _mixin


@numba.njit
def _find_distinctParent(pdg_self, pdg_all, parent_self, parent_all):
    out = parent_self.copy()
    for i in range(len(out)):
        if out[i] < 0:
            continue
        thispdg = pdg_self[i]
        parent = parent_self[i]
        parentpdg = pdg_all[parent]
        while parent >= 0 and parentpdg == thispdg:
            if parent >= len(parent_all):
                raise RuntimeError("parent index beyond length of array!")
            parent = parent_all[parent]
            parentpdg = pdg_all[parent]
        out[i] = parent

    return out


@numba.njit
def _resize(a, new_size):
    new = numpy.empty(new_size, a.dtype)
    new[:a.size] = a
    return new


@numba.njit
def _find_children(offsets_in, parentidx):
    offsets1_out = numpy.empty(len(parentidx) + 1, dtype=parentidx.dtype)
    content1_out = numpy.empty(len(parentidx), dtype=parentidx.dtype)
    offsets1_out[0] = 0

    offset0 = 0
    offset1 = 0
    for record_index in range(len(offsets_in) - 1):
        start_src, stop_src = offsets_in[record_index], offsets_in[record_index + 1]

        for index in range(stop_src - start_src):
            for possible_child in range(index, stop_src - start_src):
                if parentidx[start_src + possible_child] == index:
                    content1_out[offset1] = start_src + possible_child
                    offset1 = offset1 + 1
                    if offset1 >= len(content1_out):
                        raise RuntimeError("offset1 went out of bounds!")
            offsets1_out[offset0 + 1] = offset1
            offset0 = offset0 + 1
            if offset0 >= len(offsets1_out):
                raise RuntimeError("offset0 went out of bounds!")

    return offsets1_out, content1_out[:offset1]


class GenParticle(LorentzVector):
    '''NanoAOD generator-level particle object, including parent and child self-references

    Parent and child self-references are constructed from the ``genPartIdxMother`` column, where
    for each entry, the mother entry index is recorded, or -1 if no mother exists.
    '''
    FLAGS = [
        'isPrompt',
        'isDecayedLeptonHadron',
        'isTauDecayProduct',
        'isPromptTauDecayProduct',
        'isDirectTauDecayProduct',
        'isDirectPromptTauDecayProduct',
        'isDirectHadronDecayProduct',
        'isHardProcess',
        'fromHardProcess',
        'isHardProcessTauDecayProduct',
        'isDirectHardProcessTauDecayProduct',
        'fromHardProcessBeforeFSR',
        'isFirstCopy',
        'isLastCopy',
        'isLastCopyBeforeFSR'
    ]
    '''bit-packed statusFlags interpretations.  Use `GenParticle.hasFlags` to query'''
    enable_children = True
    'Enable construction with children column (can be turned off if causing trouble)'

    def _finalize(self, name, events):
        parent_type = awkward.type.ArrayType(float('inf'), awkward.type.OptionType(self.type.to.to))
        parent_type.check = False  # break recursion
        gen_parent = type(self)(
            self._lazy_crossref,
            args=(self._getcolumn('genPartIdxMother'), self),
            type=parent_type,
        )
        gen_parent.__doc__ = self.__doc__
        self['parent'] = gen_parent

        if self.enable_children:
            child_type = awkward.type.ArrayType(float('inf'), float('inf'), self.type.to.to)
            child_type.check = False
            children = type(self)(
                self._lazy_findchildren,
                args=(self._getcolumn('genPartIdxMother'),),
                type=child_type,
            )
            children.__doc__ = self.__doc__
            self['children'] = children

        self.type.check = False
        del self['genPartIdxMother']

    def hasFlags(self, flags):
        '''Check if one or more status flags are set

        Parameters
        ----------
            flags : list of str
                A list of flags that are required to be set true. Possible flags are enumerated in the `FLAGS` attribute

        Returns a boolean array
        '''
        dtype = self.statusFlags.type
        while not isinstance(dtype, numpy.dtype):
            try:
                dtype = dtype.to
            except AttributeError:
                dtype = dtype.type
        bits = numpy.array([self.FLAGS.index(f) for f in flags])
        mask = (1 << bits).sum().astype(dtype)
        return (self.statusFlags & mask) == mask

    @property
    def distinctParent(self):
        '''Find the particle's parent, skipping ancestors with the same pdgId'''
        array = self
        jagged = None
        if isinstance(array, awkward.VirtualArray):
            array = array.array
        if isinstance(array, awkward.JaggedArray):
            jagged = array
            array = array.content
        if isinstance(array, awkward.VirtualArray):
            array = array.array
        if isinstance(array, awkward.IndexedMaskedArray):
            mask = array.mask
            pdg_self = array.pdgId.fillna(0)
        else:
            mask = None
            pdg_self = numpy.array(array.pdgId)
        parent = array.parent
        if isinstance(array.parent, awkward.VirtualArray):
            parent = parent.array
        if not isinstance(parent, awkward.IndexedMaskedArray):
            raise RuntimeError
        parent_self = parent.mask
        if mask is not None:
            parent_self = numpy.where(mask >= 0, parent_self, -1)
        if isinstance(parent.content, awkward.VirtualArray):
            parent = parent.content.array
        pdg_all = numpy.array(parent.content.pdgId)
        parent_all = parent.content['_xref_%s_index' % self.rowname]
        globalindex = _find_distinctParent(pdg_self, pdg_all, parent_self, parent_all)
        out = type(parent)(
            globalindex,
            parent.content,
        )
        if jagged is not None:
            out = jagged.copy(content=out)
        return out

    def _lazy_findchildren(self, motherindices):
        # repair awkward type now that we've materialized
        motherindices.type.takes = self.array.offsets[-1]
        JaggedArray = self._get_mixin(self._get_methods(), awkward.JaggedArray)
        motherindices = motherindices.array
        offsets1, content1 = _find_children(
            self.array.offsets,
            motherindices.flatten()
        )
        return JaggedArray.fromoffsets(
            offsets1,
            content=self.array.content[content1]
        )


class GenVisTau(Candidate):
    '''NanoAOD visible tau object'''
    def _finalize(self, name, events):
        parent_type = awkward.type.ArrayType(float('inf'), awkward.type.OptionType(events.GenPart.type.to.to))
        parent_type.check = False  # break recursion
        gen_parent = type(events.GenPart)(
            self._lazy_crossref,
            args=(self._getcolumn('genPartIdxMother'), events.GenPart),
            type=parent_type,
        )
        gen_parent.__doc__ = self.__doc__
        self['parent'] = gen_parent
        self.type.check = False
        del self['genPartIdxMother']



"""Physics object candidate mixin

This provides just a Lorentz vector with charge, but maybe
in the future it will provide some sort of composite candiate building tool
that automatically resolves duplicates in the chain.
"""
import numpy
import awkward1
from coffea.nanoevents.methods import vector


behavior = dict(vector.behavior)


@awkward1.mixin_class(behavior)
class Candidate(vector.LorentzVector):
    """A Lorentz vector with charge

    Properties this class requires: x, y, z, t, charge
    """

    @awkward1.mixin_class_method(numpy.add, {"Candidate"})
    def add(self, other):
        return awkward1.zip(
            {
                "x": self.x + other.x,
                "y": self.y + other.y,
                "z": self.z + other.z,
                "t": self.t + other.t,
                "charge": self.charge + other.charge,
            },
            with_name="Candidate",
        )

    def sum(self, axis=-1):
        return awkward1.zip(
            {
                "x": awkward1.sum(self.x, axis=axis),
                "y": awkward1.sum(self.y, axis=axis),
                "z": awkward1.sum(self.z, axis=axis),
                "t": awkward1.sum(self.t, axis=axis),
                "charge": awkward1.sum(self.charge, axis=axis),
            },
            with_name="Candidate",
        )


@awkward1.mixin_class(behavior)
class PtEtaPhiMCandidate(Candidate, vector.PtEtaPhiMLorentzVector):
    """A Lorentz vector in eta, mass coordinates with charge

    Properties this class requires: pt, eta, phi, mass, charge
    """

    pass



from threading import Lock
import numbers
import numpy
import numba
from coffea.lookup_tools.lookup_base import lookup_base


class dense_mapped_lookup(lookup_base):
    _formulaLock = Lock()
    _formulaCache = {}

    def __init__(self, axes, mapping, formulas, feval_dim):
        self._axes = axes
        self._mapping = mapping
        self._formulas = formulas
        self._feval_dim = feval_dim

    @classmethod
    def _compile(cls, formula):
        with dense_mapped_lookup._formulaLock:
            try:
                return dense_mapped_lookup._formulaCache[formula]
            except KeyError:
                if "x" in formula:
                    feval = eval(
                        "lambda x: " + formula, {"log": numpy.log, "sqrt": numpy.sqrt}
                    )
                    out = numba.jit()(feval)
                else:
                    out = eval(formula)
                dense_mapped_lookup._formulaCache[formula] = out
                return out

    def _lookup(self, axis, values):
        if len(axis) == 2:
            return numpy.zeros(shape=values.shape, dtype=numpy.uint)
        return numpy.clip(
            numpy.searchsorted(axis, values, side="right") - 1, 0, len(axis) - 2
        )

    def _evaluate(self, *args, ignore_missing=False):
        if len(args) != len(self._axes):
            raise ValueError(
                "Incorrect number of arguments specified (expected %d got %d)"
                % (len(args), len(self._axes))
            )
        idx = (self._lookup(axis, arg) for axis, arg in zip(self._axes, args))
        mapidx = self._mapping[tuple(idx)]
        out = numpy.ones(mapidx.shape, dtype=numpy.common_type(*args))
        for ifunc in numpy.unique(mapidx):
            if ifunc < 0 and not ignore_missing:
                raise ValueError("No correction was available for some items")
            func = dense_mapped_lookup._compile(self._formulas[ifunc])
            where = mapidx == ifunc
            if isinstance(func, numbers.Number):
                out[where] = func
            else:
                if self._feval_dim is None:
                    raise RuntimeError("expected a dimension to pass to the formula")
                out[where] = func(
                    numpy.clip(
                        args[self._feval_dim][where],
                        self._axes[self._feval_dim][0],
                        self._axes[self._feval_dim][-1],
                    )
                )

        return out

    def __repr__(self):
        myrepr = "{} dimensional histogram with axes:\n".format(self._dimension)
        temp = ""
        for idim, axis in enumerate(self._axes):
            temp += "\t{}: {}\n".format(idim + 1, axis)
        myrepr += temp
        return myrepr



from abc import ABCMeta, abstractmethod
from collections import defaultdict
import numpy

try:
    from collections.abc import Set, Mapping
except ImportError:
    from collections import Set, Mapping


class AccumulatorABC(metaclass=ABCMeta):
    '''ABC for an accumulator

    Derived class must implement:
        - ``identity()``: returns a new object of same type as self,
          such that ``self + self.identity() == self``
        - ``add(other)``: adds an object of same type as self to self

    Concrete implementations are then provided for ``__add__``, ``__radd__``, and ``__iadd__``
    '''
    @abstractmethod
    def identity(self):
        '''Identity of the accumulator

        A value such that any other value added to it will return
        the other value
        '''
        pass

    @abstractmethod
    def add(self, other):
        '''Add another accumulator to this one in-place'''
        pass

    def __add__(self, other):
        ret = self.identity()
        ret.add(self)
        ret.add(other)
        return ret

    def __radd__(self, other):
        ret = self.identity()
        ret.add(other)
        ret.add(self)
        return ret

    def __iadd__(self, other):
        self.add(other)
        return self


class value_accumulator(AccumulatorABC):
    '''Holds a value of arbitrary type

    Parameters
    ----------
        default_factory : callable
            a function that returns an instance of the desired identity value
        initial : bool, optional
            an initial value, if the identity is not the desired initial value
    '''
    def __init__(self, default_factory, initial=None):
        self.value = default_factory() if initial is None else initial
        self.default_factory = default_factory

    def __repr__(self):
        if type(self.default_factory) is type:
            defrepr = self.default_factory.__name__
        else:
            defrepr = repr(self.default_factory)
        return "value_accumulator(%s, %r)" % (defrepr, self.value)

    def identity(self):
        return value_accumulator(self.default_factory)

    def add(self, other):
        if isinstance(other, value_accumulator):
            self.value = self.value + other.value
        else:
            self.value = self.value + other


class list_accumulator(list, AccumulatorABC):
    '''A list with accumulator semantics

    See `list` for further info
    '''
    def identity(self):
        return list()

    def add(self, other):
        '''Add another accumulator to this one in-place'''
        if isinstance(other, list):
            list.extend(self, other)
        else:
            raise ValueError


class set_accumulator(set, AccumulatorABC):
    '''A set with accumulator semantics

    See `set` for further info
    '''
    def identity(self):
        return set_accumulator()

    def add(self, other):
        '''Add another accumulator to this one in-place

        Note
        ----
        This replaces `set.add` behavior, unfortunately.
        A workaround is to use `set.update`, e.g. ``a.update({'val'})``
        '''
        if isinstance(other, Set):
            set.update(self, other)
        else:
            set.add(self, other)


class dict_accumulator(dict, AccumulatorABC):
    '''A dictionary with accumulator semantics

    See `dict` for further info.
    It is assumed that the contents of the dict have accumulator semantics.
    '''
    def identity(self):
        ret = dict_accumulator()
        for key, value in self.items():
            ret[key] = value.identity()
        return ret

    def add(self, other):
        if isinstance(other, Mapping):
            for key, value in other.items():
                if key not in self:
                    if isinstance(value, AccumulatorABC):
                        self[key] = value.identity()
                    else:
                        raise ValueError
                self[key] += value
        else:
            raise ValueError


class defaultdict_accumulator(defaultdict, AccumulatorABC):
    '''A defaultdict with accumulator semantics

    See `collections.defaultdict` for further info.
    It is assumed that the contents of the dict have accumulator semantics
    '''
    def identity(self):
        return defaultdict_accumulator(self.default_factory)

    def add(self, other):
        for key, value in other.items():
            self[key] += value


class column_accumulator(AccumulatorABC):
    '''An appendable numpy ndarray

    Parameters
    ----------
        value : numpy.ndarray
            The identity value array, which should be an empty ndarray
            with the desired row shape. The column dimension will correspond to
            the first index of `value` shape.

    Examples
    --------
    If a set of accumulators is defined as::

        a = column_accumulator(np.array([]))
        b = column_accumulator(np.array([1., 2., 3.]))
        c = column_accumulator(np.array([4., 5., 6.]))

    then:

    >>> a + b
    column_accumulator(array([1., 2., 3.]))
    >>> c + b + a
    column_accumulator(array([4., 5., 6., 1., 2., 3.]))
    '''
    def __init__(self, value):
        if not isinstance(value, numpy.ndarray):
            raise ValueError("column_accumulator only works with numpy arrays")
        self._empty = numpy.zeros(dtype=value.dtype, shape=(0,) + value.shape[1:])
        self._value = value

    def __repr__(self):
        return "column_accumulator(%r)" % self.value

    def identity(self):
        return column_accumulator(self._empty)

    def add(self, other):
        if not isinstance(other, column_accumulator):
            raise ValueError("column_accumulator cannot be added to %r" % type(other))
        if other._empty.shape != self._empty.shape:
            raise ValueError("Cannot add two column_accumulator objects of dissimilar shape (%r vs %r)"
                             % (self._empty.shape, other._empty.shape))
        self._value = numpy.concatenate((self._value, other._value))

    @property
    def value(self):
        '''The current value of the column

        Returns a numpy array where the first dimension is the column dimension
        '''
        return self._value



from coffea import hist, processor


class NanoEvents0Processor(processor.ProcessorABC):
    def __init__(self, columns=[], canaries=[]):
        self._columns = columns
        self._canaries = canaries
        dataset_axis = hist.Cat("dataset", "Primary dataset")
        mass_axis = hist.Bin("mass", r"$m_{\mu\mu}$ [GeV]", 30000, 0.25, 300)
        pt_axis = hist.Bin("pt", r"$p_{T}$ [GeV]", 30000, 0.25, 300)

        self._accumulator = processor.dict_accumulator(
            {
                'mass': hist.Hist("Counts", dataset_axis, mass_axis),
                'pt': hist.Hist("Counts", dataset_axis, pt_axis),
                'cutflow': processor.defaultdict_accumulator(int),
                'worker': processor.set_accumulator(),
            }
        )

    @property
    def columns(self):
        return self._columns

    @property
    def accumulator(self):
        return self._accumulator

    def process(self, events):
        output = self.accumulator.identity()

        dataset = events.metadata['dataset']

        dimuon = events.Muon.choose(2)
        dimuon = dimuon.i0 + dimuon.i1

        magickey = events.Muon.array.content.pt.persistentkey
        if magickey in self._canaries:
            try:
                from distributed import get_worker
                worker = get_worker()
                output['worker'].add(worker.name)
            except ValueError:
                pass

        output['pt'].fill(dataset=dataset, pt=events.Muon.pt.flatten())
        output['mass'].fill(dataset=dataset, mass=dimuon.mass.flatten())
        output['cutflow']['%s_pt' % dataset] += sum(events.Muon.counts)
        output['cutflow']['%s_mass' % dataset] += sum(dimuon.counts)

        return output

    def postprocess(self, accumulator):
        return accumulator



from coffea.nanoevents import transforms
from coffea.nanoevents.util import quote, concat


def listarray_form(content, offsets):
    if offsets["class"] != "NumpyArray":
        raise ValueError
    if offsets["primitive"] == "int32":
        arrayclass = "ListOffsetArray32"
        offsetstype = "i32"
    elif offsets["primitive"] == "int64":
        arrayclass = "ListOffsetArray64"
        offsetstype = "i64"
    else:
        raise ValueError("Unrecognized offsets data type")
    return {
        "class": arrayclass,
        "offsets": offsetstype,
        "content": content,
        "form_key": concat(offsets["form_key"], "!skip"),
    }


def zip_forms(forms, name, record_name=None, offsets=None):
    if not isinstance(forms, dict):
        raise ValueError("Expected a dictionary")
    if all(form["class"].startswith("ListOffsetArray") for form in forms.values()):
        first = next(iter(forms.values()))
        if not all(form["class"] == first["class"] for form in forms.values()):
            raise ValueError
        if not all(form["offsets"] == first["offsets"] for form in forms.values()):
            raise ValueError
        record = {
            "class": "RecordArray",
            "contents": {k: form["content"] for k, form in forms.items()},
            "form_key": quote("!invalid," + name),
        }
        if record_name is not None:
            record["parameters"] = {"__record__": record_name}
        if offsets is None:
            return {
                "class": first["class"],
                "offsets": first["offsets"],
                "content": record,
                "form_key": first["form_key"],
            }
        else:
            return listarray_form(record, offsets)
    elif all(form["class"] == "NumpyArray" for form in forms.values()):
        record = {
            "class": "RecordArray",
            "contents": {k: form for k, form in forms.items()},
            "form_key": quote("!invalid," + name),
        }
        if record_name is not None:
            record["parameters"] = {"__record__": record_name}
        return record
    else:
        raise NotImplementedError("Cannot zip forms")


def nest_jagged_forms(parent, child, counts_name, name):
    """Place child listarray inside parent listarray as a double-jagged array"""
    if not parent["class"].startswith("ListOffsetArray"):
        raise ValueError
    if parent["content"]["class"] != "RecordArray":
        raise ValueError
    if not child["class"].startswith("ListOffsetArray"):
        raise ValueError
    counts = parent["content"]["contents"][counts_name]
    offsets = transforms.counts2offsets_form(counts)
    inner = listarray_form(child["content"], offsets)
    parent["content"]["contents"][name] = inner


class BaseSchema:
    """Base schema builder

    The basic schema is essentially unchanged from the original ROOT file.
    A top-level `base.NanoEvents` object is returned, where each original branch
    form is accessible as a direct descendant.
    """

    behavior = {}

    def __init__(self, base_form):
        params = dict(base_form.get("parameters", {}))
        params["__record__"] = "NanoEvents"
        params.setdefault("metadata", {})
        self._form = {
            "class": "RecordArray",
            "contents": base_form["contents"],
            "parameters": params,
            "form_key": "",
        }

    @property
    def form(self):
        """Awkward1 form of this schema"""
        return self._form

    @property
    def behavior(self):
        """Behaviors necessary to implement this schema"""
        from coffea.nanoevents.methods import base

        return base.behavior


class NanoAODSchema(BaseSchema):
    """NanoAOD schema builder

    The NanoAOD schema is built from all branches found in the supplied file, based on
    the naming pattern of the branches. The following additional arrays are constructed:

    - Any branches named ``n{name}`` are assumed to be counts branches and converted to offsets ``o{name}``
    - Any local index branches with names matching ``{source}_{target}Idx*`` are converted to global indexes for the event chunk (postfix ``G``)
    - Any `nested_items` are constructed, if the necessary branches are available
    - Any `special_items` are constructed, if the necessary branches are available

    From those arrays, NanoAOD collections are formed as collections of branches grouped by name, where:

    - one branch exists named ``name`` and no branches start with ``name_``, interpreted as a single flat array;
    - one branch exists named ``name``, one named ``n{name}``, and no branches start with ``name_``, interpreted as a single jagged array;
    - no branch exists named ``{name}`` and many branches start with ``name_*``, interpreted as a flat table; or
    - one branch exists named ``n{name}`` and many branches start with ``name_*``, interpreted as a jagged table.

    Collections are assigned mixin types according to the `mixins` mapping.
    All collections are then zipped into one `base.NanoEvents` record and returned.
    """

    mixins = {
        "CaloMET": "MissingET",
        "ChsMET": "MissingET",
        "GenMET": "MissingET",
        "MET": "MissingET",
        "METFixEE2017": "MissingET",
        "PuppiMET": "MissingET",
        "RawMET": "MissingET",
        "TkMET": "MissingET",
        # pseudo-lorentz: pt, eta, phi, mass=0
        "IsoTrack": "PtEtaPhiMCollection",
        "SoftActivityJet": "PtEtaPhiMCollection",
        "TrigObj": "PtEtaPhiMCollection",
        # True lorentz: pt, eta, phi, mass
        "FatJet": "FatJet",
        "GenDressedLepton": "PtEtaPhiMCollection",
        "GenJet": "PtEtaPhiMCollection",
        "GenJetAK8": "FatJet",
        "Jet": "Jet",
        "LHEPart": "PtEtaPhiMCollection",
        "SV": "PtEtaPhiMCollection",
        "SubGenJetAK8": "PtEtaPhiMCollection",
        "SubJet": "PtEtaPhiMCollection",
        # Candidate: lorentz + charge
        "Electron": "Electron",
        "Muon": "Muon",
        "Photon": "Photon",
        "FsrPhoton": "FsrPhoton",
        "Tau": "Tau",
        "GenVisTau": "GenVisTau",
        # special
        "GenPart": "GenParticle",
    }
    """Default configuration for mixin types, based on the collection name.

    The types are implemented in the `coffea.nanoevents.methods.nanoaod` module.
    """
    nested_items = {
        "FatJet_subJetIdxG": ["FatJet_subJetIdx1G", "FatJet_subJetIdx2G"],
        "Jet_muonIdxG": ["Jet_muonIdx1G", "Jet_muonIdx2G"],
        "Jet_electronIdxG": ["Jet_electronIdx1G", "Jet_electronIdx2G"],
    }
    """Default nested collections, where nesting is accomplished by a fixed-length set of indexers"""
    special_items = {
        "GenPart_distinctParentIdxG": (
            transforms.distinctParent_form,
            ("GenPart_genPartIdxMotherG", "GenPart_pdgId"),
        ),
        "GenPart_childrenIdxG": (
            transforms.children_form,
            (
                "oGenPart",
                "GenPart_genPartIdxMotherG",
            ),
        ),
        "GenPart_distinctChildrenIdxG": (
            transforms.children_form,
            (
                "oGenPart",
                "GenPart_distinctParentIdxG",
            ),
        ),
    }
    """Default special arrays, where the callable and input arrays are specified in the value"""

    def __init__(self, base_form, version="6"):
        self._version = version
        super().__init__(base_form)
        self._form["contents"] = self._build_collections(self._form["contents"])
        self._form["parameters"]["metadata"]["version"] = self._version

    def _build_collections(self, branch_forms):
        # parse into high-level records (collections, list collections, and singletons)
        collections = set(k.split("_")[0] for k in branch_forms)
        collections -= set(
            k for k in collections if k.startswith("n") and k[1:] in collections
        )

        # Create offsets virtual arrays
        for name in collections:
            if "n" + name in branch_forms:
                branch_forms["o" + name] = transforms.counts2offsets_form(
                    branch_forms["n" + name]
                )

        # Create global index virtual arrays for indirection
        idxbranches = [k for k in branch_forms if "Idx" in k]
        for name in collections:
            indexers = [k for k in idxbranches if k.startswith(name)]
            for k in indexers:
                target = k[len(name) + 1 : k.find("Idx")]
                target = target[0].upper() + target[1:]
                if target not in collections:
                    raise RuntimeError(
                        "Parsing indexer %s, expected to find collection %s but did not"
                        % (k, target)
                    )
                branch_forms[k + "G"] = transforms.local2global_form(
                    branch_forms[k], branch_forms["o" + target]
                )

        # Create nested indexer from Idx1, Idx2, ... arrays
        for name, indexers in self.nested_items.items():
            if all(idx in branch_forms for idx in indexers):
                branch_forms[name] = transforms.nestedindex_form(
                    [branch_forms[idx] for idx in indexers]
                )

        # Create any special arrays
        for name, (fcn, args) in self.special_items.items():
            if all(k in branch_forms for k in args):
                branch_forms[name] = fcn(*(branch_forms[k] for k in args))

        output = {}
        for name in collections:
            mixin = self.mixins.get(name, "NanoCollection")
            if "o" + name in branch_forms and name not in branch_forms:
                # list collection
                offsets = branch_forms["o" + name]
                content = {
                    k[len(name) + 1 :]: branch_forms[k]
                    for k in branch_forms
                    if k.startswith(name + "_")
                }
                output[name] = zip_forms(
                    content, name, record_name=mixin, offsets=offsets
                )
                output[name]["content"]["parameters"].update(
                    {
                        "__doc__": offsets["parameters"]["__doc__"],
                        "collection_name": name,
                    }
                )
            elif "o" + name in branch_forms:
                # list singleton, can use branch's own offsets
                output[name] = branch_forms[name]
                output[name]["parameters"].update(
                    {"__array__": mixin, "collection_name": name}
                )
            elif name in branch_forms:
                # singleton
                output[name] = branch_forms[name]
            else:
                # simple collection
                output[name] = zip_forms(
                    {
                        k[len(name) + 1 :]: branch_forms[k]
                        for k in branch_forms
                        if k.startswith(name + "_")
                    },
                    name,
                    record_name=mixin,
                )
                output[name]["parameters"].update({"collection_name": name})

        return output

    @property
    def behavior(self):
        """Behaviors necessary to implement this schema"""
        from coffea.nanoevents.methods import nanoaod

        return nanoaod.behavior


class TreeMakerSchema(BaseSchema):
    """TreeMaker schema builder

    The TreeMaker schema is built from all branches found in the supplied file, based on
    the naming pattern of the branches. From those arrays, TreeMaker collections are formed
    as collections of branches grouped by name, where:

    FIX ME

    All collections are then zipped into one `base.NanoEvents` record and returned.
    """

    def __init__(self, base_form):
        super().__init__(base_form)
        self._form["contents"] = self._build_collections(self._form["contents"])

    def _build_collections(self, branch_forms):
        # Turn any special classes into the appropriate awkward form
        composite_objects = list(set(k.split("/")[0] for k in branch_forms if "/" in k))
        for objname in composite_objects:
            # grab the * from "objname/objname.*"
            components = set(
                k[2 * len(objname) + 2 :]
                for k in branch_forms
                if k.startswith(objname + "/")
            )
            if components == {
                "fCoordinates.fPt",
                "fCoordinates.fEta",
                "fCoordinates.fPhi",
                "fCoordinates.fE",
            }:
                form = zip_forms(
                    {
                        "pt": branch_forms.pop(f"{objname}/{objname}.fCoordinates.fPt"),
                        "eta": branch_forms.pop(
                            f"{objname}/{objname}.fCoordinates.fEta"
                        ),
                        "phi": branch_forms.pop(
                            f"{objname}/{objname}.fCoordinates.fPhi"
                        ),
                        "energy": branch_forms.pop(
                            f"{objname}/{objname}.fCoordinates.fE"
                        ),
                    },
                    objname,
                    "PtEtaPhiELorentzVector",
                )
                print("PtEtaPhiELorentzVector:", objname)
                branch_forms[objname] = form
            elif components == {
                "fCoordinates.fX",
                "fCoordinates.fY",
                "fCoordinates.fZ",
            }:
                form = zip_forms(
                    {
                        "x": branch_forms.pop(f"{objname}/{objname}.fCoordinates.fX"),
                        "y": branch_forms.pop(f"{objname}/{objname}.fCoordinates.fY"),
                        "z": branch_forms.pop(f"{objname}/{objname}.fCoordinates.fZ"),
                    },
                    objname,
                    "Point",
                )
                print("Point:", objname)
                branch_forms[objname] = form
            else:
                raise ValueError(
                    f"Unrecognized class with split branches: {components}"
                )

        collections = [
            "Electrons",
            "GenElectrons",
            "GenJets",
            "GenJetsAK8",
            "GenMuons",
            "GenParticles",
            "GenTaus",
            "GenVertices",
            "Jets",
            "JetsAK8",
            "JetsAK8_subjets",
            "JetsAK15",
            "Muons",
            "PrimaryVertices",
            "TAPElectronTracks",
            "TAPMuonTracks",
            "TAPPionTracks",
            "Tracks",
        ]
        for cname in collections:
            items = sorted(k for k in branch_forms if k.startswith(cname + "_"))
            if len(items) == 0:
                continue
            if cname == "JetsAK8":
                items = [k for k in items if not k.startswith("JetsAK8_subjets")]
                items.append("JetsAK8_subjetsCounts")
            if cname == "JetsAK8_subjets":
                items = [k for k in items if not k.endswith("Counts")]
            if cname not in branch_forms:
                collection = zip_forms(
                    {k[len(cname) + 1]: branch_forms.pop(k) for k in items}, cname
                )
                branch_forms[cname] = collection
            else:
                collection = branch_forms[cname]
                if not collection["class"].startswith("ListOffsetArray"):
                    raise NotImplementedError(
                        f"{cname} isn't a jagged array, not sure what to do"
                    )
                for item in items:
                    itemname = item[len(cname) + 1 :]
                    collection["content"]["contents"][itemname] = branch_forms.pop(
                        item
                    )["content"]

        nest_jagged_forms(
            branch_forms["JetsAK8"],
            branch_forms.pop("JetsAK8_subjets"),
            "subjetsCounts",
            "subjets",
        )

        return branch_forms

    @property
    def behavior(self):
        """Behaviors necessary to implement this schema"""
        from coffea.nanoevents.methods import base, vector

        behavior = {}
        behavior.update(base.behavior)
        behavior.update(vector.behavior)
        return behavior



import copy
import numpy
import numba
import awkward1
from coffea.nanoevents.util import concat


def offsets(stack):
    """Extract offsets from ListOffsetArray

    Signature: array,!offsets
    """
    stack.append(stack.pop().layout.offsets)


def mask(stack):
    """Extract mask from a masked array

    Signature: array,!mask
    """
    stack.append(stack.pop().layout.mask)


def index(stack):
    """Extract index from array

    Signature: array,!index
    """
    stack.append(stack.pop().layout.index)


def starts(stack):
    """Extract offsets from ListArray

    Signature: array,!offsets
    """
    stack.append(stack.pop().layout.starts)


def stops(stack):
    """Extract offsets from ListArray

    Signature: array,!stops
    """
    stack.append(stack.pop().layout.stops)


def tags(stack):
    """Extract tags from UnionArray

    Signature: array,!tags
    """
    stack.append(stack.pop().layout.tags)


def content(stack):
    """Extract content from array

    Signature: array,!content
    """
    stack.append(stack.pop().layout.content)


def counts2offsets_form(counts_form):
    form = {
        "class": "NumpyArray",
        "itemsize": 8,
        "format": "i",
        "primitive": "int64",
        "parameters": counts_form.get("parameters", None),
        "form_key": concat(counts_form["form_key"], "!counts2offsets"),
    }
    return form


def counts2offsets(stack):
    """Cumulative sum of counts

    Signature: counts,!counts2offsets
    Outputs an array with length one larger than input
    """
    counts = numpy.array(stack.pop())
    offsets = numpy.empty(len(counts) + 1, dtype=numpy.int64)
    offsets[0] = 0
    numpy.cumsum(counts, out=offsets[1:])
    stack.append(offsets)


def local2global_form(index, target_offsets):
    if not index["class"].startswith("ListOffset"):
        raise RuntimeError
    if not target_offsets["class"] == "NumpyArray":
        raise RuntimeError
    form = copy.deepcopy(index)
    form["content"]["form_key"] = concat(
        index["form_key"], target_offsets["form_key"], "!local2global"
    )
    form["content"]["itemsize"] = 8
    form["content"]["primitive"] = "int64"
    return form


def local2global(stack):
    """Turn jagged local index into global index

    Signature: index,target_offsets,!local2global
    Outputs a content array with same shape as index content
    """
    target_offsets = stack.pop()
    index = stack.pop()
    index = index.mask[index >= 0] + target_offsets[:-1]
    out = numpy.array(awkward1.flatten(awkward1.fill_none(index, -1)))
    if out.dtype != numpy.int64:
        raise RuntimeError
    stack.append(out)


@numba.njit
def _distinctParent_kernel(allpart_parent, allpart_pdg):
    out = numpy.empty(len(allpart_pdg), dtype=numpy.int64)
    for i in range(len(allpart_pdg)):
        parent = allpart_parent[i]
        if parent < 0:
            out[i] = -1
            continue
        thispdg = allpart_pdg[i]
        while parent >= 0 and allpart_pdg[parent] == thispdg:
            if parent >= len(allpart_pdg):
                raise RuntimeError("parent index beyond length of array!")
            parent = allpart_parent[parent]
        out[i] = parent
    return out


def distinctParent_form(parents, pdg):
    if not parents["class"].startswith("ListOffset"):
        raise RuntimeError
    if not pdg["class"].startswith("ListOffset"):
        raise RuntimeError
    form = {
        "class": "ListOffsetArray64",
        "offsets": "i64",
        "content": {
            "class": "NumpyArray",
            "itemsize": 8,
            "format": "i",
            "primitive": "int64",
        },
        "form_key": parents["form_key"],
    }
    form["content"]["form_key"] = concat(
        parents["content"]["form_key"],
        pdg["content"]["form_key"],
        "!distinctParent",
    )
    return form


def distinctParent(stack):
    """Compute first parent with distinct PDG id

    Signature: globalparents,globalpdgs,!distinctParent
    Expects global indexes, flat arrays, which should be same length
    """
    pdg = stack.pop()
    parents = stack.pop()
    stack.append(_distinctParent_kernel(awkward1.Array(parents), awkward1.Array(pdg)))


@numba.njit
def _children_kernel(offsets_in, parentidx):
    offsets1_out = numpy.empty(len(parentidx) + 1, dtype=numpy.int64)
    content1_out = numpy.empty(len(parentidx), dtype=numpy.int64)
    offsets1_out[0] = 0

    offset0 = 0
    offset1 = 0
    for record_index in range(len(offsets_in) - 1):
        start_src, stop_src = offsets_in[record_index], offsets_in[record_index + 1]

        for index in range(start_src, stop_src):
            for possible_child in range(index, stop_src):
                if parentidx[possible_child] == index:
                    content1_out[offset1] = possible_child
                    offset1 = offset1 + 1
                    if offset1 >= len(content1_out):
                        raise RuntimeError("offset1 went out of bounds!")
            offsets1_out[offset0 + 1] = offset1
            offset0 = offset0 + 1
            if offset0 >= len(offsets1_out):
                raise RuntimeError("offset0 went out of bounds!")

    return offsets1_out, content1_out[:offset1]


def children_form(offsets, globalparents):
    if not globalparents["class"].startswith("ListOffset"):
        raise RuntimeError
    form = {
        "class": "ListOffsetArray64",
        "offsets": "i64",
        "content": {
            "class": "ListOffsetArray64",
            "offsets": "i64",
            "content": {
                "class": "NumpyArray",
                "itemsize": 8,
                "format": "i",
                "primitive": "int64",
            },
        },
    }
    form["form_key"] = offsets["form_key"]
    key = concat(offsets["form_key"], globalparents["content"]["form_key"], "!children")
    form["content"]["form_key"] = key
    form["content"]["content"]["form_key"] = concat(key, "!content")
    return form


def children(stack):
    """Compute children

    Signature: offsets,globalparents,!children
    Output will be a jagged array with same outer shape as globalparents content
    """
    parents = stack.pop()
    offsets = stack.pop()
    coffsets, ccontent = _children_kernel(offsets, parents)
    out = awkward1.Array(
        awkward1.layout.ListOffsetArray64(
            awkward1.layout.Index64(coffsets),
            awkward1.layout.NumpyArray(ccontent),
        )
    )
    stack.append(out)


def nestedindex_form(indices):
    if not all(index["class"].startswith("ListOffset") for index in indices):
        raise RuntimeError
    form = {
        "class": "ListOffsetArray64",
        "offsets": "i64",
        "content": copy.deepcopy(indices[0]),
    }
    # steal offsets from first input
    key = []
    for index in indices:
        key.append(index["content"]["form_key"])
    key.append("!nestedindex")
    key = concat(*key)
    form["form_key"] = indices[0]["form_key"]
    form["content"]["form_key"] = key
    form["content"]["content"]["form_key"] = concat(key, "!content")
    return form


def nestedindex(stack):
    """Concatenate a list of indices along a new axis

    Signature: index1,index2,...,!nestedindex
    Index arrays should all be same shape flat arrays
    Outputs a jagged array with same outer shape as index arrays
    """
    indexers = stack[:]
    stack.clear()
    # return awkward1.concatenate([idx[:, None] for idx in indexers], axis=1)
    n = len(indexers)
    out = numpy.empty(n * len(indexers[0]), dtype="int64")
    for i, idx in enumerate(indexers):
        out[i::n] = idx
    offsets = numpy.arange(0, len(out) + 1, n, dtype=numpy.int64)
    out = awkward1.Array(
        awkward1.layout.ListOffsetArray64(
            awkward1.layout.Index64(offsets),
            awkward1.layout.NumpyArray(out),
        )
    )
    stack.append(out)



from coffea import hist, processor
from copy import deepcopy
from concurrent.futures import ThreadPoolExecutor, as_completed

from tqdm import tqdm
import pickle as pkl
import lz4.frame as lz4f
import numpy as np
import pandas as pd
from functools import partial

from ..executor import _futures_handler
from coffea.nanoaod import NanoEvents
from coffea.nanoevents import schemas

import pyspark
import pyspark.sql.functions as fn
from pyspark.sql.types import BinaryType, StringType, StructType, StructField

from jinja2 import Environment, PackageLoader, select_autoescape
from coffea.util import awkward

lz4_clevel = 1


# this is a UDF that takes care of summing histograms across
# various spark results where the outputs are histogram blobs
def agg_histos_raw(series, processor_instance, lz4_clevel):
    goodlines = series[series.str.len() > 0]
    if goodlines.size == 1:  # short-circuit trivial aggregations
        return goodlines[0]
    outhist = processor_instance.accumulator.identity()
    for line in goodlines:
        outhist.add(pkl.loads(lz4f.decompress(line)))
    return lz4f.compress(pkl.dumps(outhist), compression_level=lz4_clevel)


@fn.pandas_udf(BinaryType(), fn.PandasUDFType.GROUPED_AGG)
def agg_histos(series):
    global processor_instance, lz4_clevel
    return agg_histos_raw(series, processor_instance, lz4_clevel)


def reduce_histos_raw(df, processor_instance, lz4_clevel):
    histos = df['histos']
    mask = (histos.str.len() > 0)
    outhist = processor_instance.accumulator.identity()
    for line in histos[mask]:
        outhist.add(pkl.loads(lz4f.decompress(line)))
    return pd.DataFrame(data={'histos': np.array([lz4f.compress(pkl.dumps(outhist), compression_level=lz4_clevel)], dtype='O')})


@fn.pandas_udf(StructType([StructField('histos', BinaryType(), True)]), fn.PandasUDFType.GROUPED_MAP)
def reduce_histos(df):
    global processor_instance, lz4_clevel
    return reduce_histos_raw(df, processor_instance, lz4_clevel)


class SparkExecutor(object):
    _template_name = 'spark.py.tmpl'

    def __init__(self):
        self._cacheddfs = None
        self._rawresults = None
        self._counts = None
        self._env = Environment(loader=PackageLoader('coffea.processor',
                                                     'templates'),
                                autoescape=select_autoescape(['py'])
                                )

    @property
    def counts(self):
        return self._counts

    def __call__(self, spark, dfslist, theprocessor, output, thread_workers,
                 use_df_cache, flatten, schema, status=True, unit='datasets', desc='Processing'):
        # processor needs to be a global
        global processor_instance, coffea_udf, coffea_udf_flat, coffea_udf_nano
        processor_instance = theprocessor
        # get columns from processor
        columns = processor_instance.columns
        cols_w_ds = ['dataset'] + columns
        # make our udf
        tmpl = self._env.get_template(self._template_name)
        render = tmpl.render(cols=columns)
        exec(render)

        # cache the input datasets if it's not already done
        if self._counts is None:
            self._counts = {}
            # go through each dataset and thin down to the columns we want
            for ds, (df, counts) in dfslist.items():
                self._counts[ds] = counts

        def spex_accumulator(total, result):
            ds, df = result
            total[ds] = df

        if self._cacheddfs is None:
            self._cacheddfs = {}
            cachedesc = 'caching' if use_df_cache else 'pruning'
            with ThreadPoolExecutor(max_workers=thread_workers) as executor:
                futures = set()
                for ds, (df, counts) in dfslist.items():
                    futures.add(executor.submit(self._pruneandcache_data, ds, df, cols_w_ds, use_df_cache))
                _futures_handler(futures, self._cacheddfs, status, unit, cachedesc, spex_accumulator, None)

        with ThreadPoolExecutor(max_workers=thread_workers) as executor:
            futures = set()
            for ds, df in self._cacheddfs.items():
                co_udf = coffea_udf
                if flatten:
                    co_udf = coffea_udf_flat
                if schema is not None:
                    if schema is NanoEvents:
                        co_udf = coffea_udf_nano
                    elif issubclass(schema, schemas.BaseSchema):
                        raise NotImplementedError("The uproot4 version of NanoEvents (%s) has not been implemented yet" % (str(schema.__name__)))
                    else:
                        raise ValueError("Expected schema to derive from BaseSchema or be an instance of NanoEvents (%s)" % (str(schema.__name__)))
                futures.add(executor.submit(self._launch_analysis, ds, df, co_udf, cols_w_ds))
            # wait for the spark jobs to come in
            self._rawresults = {}
            _futures_handler(futures, self._rawresults, status, unit, desc, spex_accumulator, None)

        for ds, bitstream in self._rawresults.items():
            if bitstream is None:
                raise Exception('No pandas dataframe returned from spark in dataset: %s, something went wrong!' % ds)
            if bitstream.empty:
                raise Exception('The histogram list returned from spark is empty in dataset: %s, something went wrong!' % ds)
            bits = bitstream[bitstream.columns[0]][0]
            output.add(pkl.loads(lz4f.decompress(bits)))

    def _pruneandcache_data(self, ds, df, columns, cacheit):
        if cacheit:
            return ds, df.select(*columns).cache()
        return ds, df.select(*columns)

    def _launch_analysis(self, ds, df, udf, columns):
        histo_map_parts = (df.rdd.getNumPartitions() // 20) + 1
        return ds, df.select(udf(*columns).alias('histos')) \
                     .withColumn('hpid', fn.spark_partition_id() % histo_map_parts) \
                     .repartition(histo_map_parts, 'hpid') \
                     .groupBy('hpid').apply(reduce_histos) \
                     .groupBy().agg(agg_histos('histos')) \
                     .toPandas()


spark_executor = SparkExecutor()



import awkward


def _mixin(methods, awkwardtype):
    '''Like awkward.Methods.mixin but also captures methods in dir() and propagate docstr'''
    if not issubclass(awkwardtype, awkward.array.base.AwkwardArray):
        raise ValueError("Only mix in to awkward types or derived")
    if not issubclass(methods, awkward.array.objects.Methods):
        raise ValueError("Can only mixin methods deriving from awkward Methods ABC")
    newtype = type(methods.__name__ + 'Array', (methods, awkwardtype), {})
    newtype.__dir__ = lambda self: dir(methods) + awkwardtype.__dir__(self)
    newtype.__doc__ = methods.__doc__
    return newtype



import numpy
import awkward
import awkward1
import numbers


class lookup_base(object):
    """Base class for all objects that do some sort of value or function lookup"""

    def __init__(self):
        pass

    def __call__(self, *args, **kwargs):
        if all(isinstance(x, (numpy.ndarray, numbers.Number)) for x in args):
            return self._evaluate(*args, **kwargs)
        elif any(isinstance(x, awkward.JaggedArray) for x in args):
            return self._call_ak0(list(args), **kwargs)

        def getfunction(inputs, depth):
            if all(
                isinstance(x, awkward1.layout.NumpyArray)
                or not isinstance(
                    x, (awkward1.layout.Content, awkward1.partition.PartitionedArray)
                )
                for x in inputs
            ):
                nplike = awkward1.nplike.of(*inputs)
                if not isinstance(nplike, awkward1.nplike.Numpy):
                    raise NotImplementedError(
                        "support for cupy/jax/etc. numpy extensions"
                    )
                result = self._evaluate(*[nplike.asarray(x) for x in inputs], **kwargs)
                return lambda: (awkward1.layout.NumpyArray(result),)
            return None

        behavior = awkward1._util.behaviorof(*args)
        args = [
            awkward1.operations.convert.to_layout(
                arg, allow_record=False, allow_other=True
            )
            for arg in args
        ]
        out = awkward1._util.broadcast_and_apply(args, getfunction, behavior)
        assert isinstance(out, tuple) and len(out) == 1
        return awkward1._util.wrap(out[0], behavior)

    def _call_ak0(self, inputs, **kwargs):
        offsets = None
        # TODO: check can use offsets (this should always be true for striped)
        # Alternatively we can just use starts and stops
        for i in range(len(inputs)):
            if isinstance(inputs[i], awkward.JaggedArray):
                if offsets is not None and offsets.base is not inputs[i].offsets.base:
                    if type(offsets) is int:
                        raise Exception(
                            "Do not mix JaggedArrays and numpy arrays when calling derived class of lookup_base"
                        )
                    elif (
                        type(offsets) is numpy.ndarray
                        and offsets.base is not inputs[i].offsets.base
                    ):
                        raise Exception(
                            "All input jagged arrays must have a common structure (offsets)!"
                        )
                offsets = inputs[i].offsets
                inputs[i] = inputs[i].content
            elif isinstance(inputs[i], numpy.ndarray):
                if offsets is not None:
                    if type(offsets) is numpy.ndarray:
                        raise Exception(
                            "do not mix JaggedArrays and numpy arrays when calling a derived class of lookup_base"
                        )
                offsets = -1
        retval = self._evaluate(*tuple(inputs), **kwargs)
        if offsets is not None and type(offsets) is not int:
            retval = awkward.JaggedArray.fromoffsets(offsets, retval)
        return retval

    def _evaluate(self, *args, **kwargs):
        raise NotImplementedError



"""2D, 3D, and Lorentz vector class mixins

These mixins will eventually be superceded by the `vector <https://github.com/scikit-hep/vector>`__ library,
which will hopefully be feature-compatible.

A small example::

    import numpy as np
    import awkward1 as ak
    from coffea.nanoevents.methods import vector
    ak.behavior.update(vector.behavior)

    n = 1000

    vec = ak.zip(
        {
            "x": np.random.normal(size=n),
            "y": np.random.normal(size=n),
            "z": np.random.normal(size=n),
        },
        with_name="ThreeVector",
    )

    vec4 = ak.zip(
        {
            "pt": vec.r,
            "eta": -np.log(np.tan(vec.theta/2)),
            "phi": vec.phi,
            "mass": np.full(n, 1.),
        },
        with_name="PtEtaPhiMLorentzVector",
    )

    assert np.allclose(np.array(vec4.x), np.array(vec.x))
    assert np.allclose(np.array(vec4.y), np.array(vec.y))
    assert np.allclose(np.array(vec4.z), np.array(vec.z))
    assert np.allclose(np.array(abs(2*vec + vec4) / abs(vec)), 3)

"""
import numbers
import numpy
import awkward1


behavior = {}


@awkward1.mixin_class(behavior)
class TwoVector:
    """A cartesian 2-dimensional vector

    A heavy emphasis towards a momentum vector interpretation is assumed.
    Properties this class requires: x, y
    """

    @property
    def r(self):
        return numpy.sqrt(self.r2)

    @property
    def phi(self):
        return numpy.arctan2(self.y, self.x)

    @property
    def px(self):
        return self.x

    @property
    def py(self):
        return self.y

    @property
    def r2(self):
        return self.x ** 2 + self.y ** 2

    @property
    def pt2(self):
        return self.r2

    @property
    def pt(self):
        return self.r

    @awkward1.mixin_class_method(numpy.absolute)
    def abs(self):
        return self.r

    @awkward1.mixin_class_method(numpy.add, {"TwoVector"})
    def add(self, other):
        return awkward1.zip(
            {"x": self.x + other.x, "y": self.y + other.y},
            with_name="TwoVector",
        )

    def sum(self, axis=-1):
        out = awkward1.zip(
            {
                "x": awkward1.sum(self.x, axis=axis),
                "y": awkward1.sum(self.y, axis=axis),
            },
            with_name="TwoVector",
            highlevel=False,
        )
        return awkward1._util.wrap(out, cache=self.cache, behavior=self.behavior)

    @awkward1.mixin_class_method(numpy.multiply, {numbers.Number})
    def prod(self, other):
        return awkward1.zip(
            {"x": self.x * other, "y": self.y * other},
            with_name="TwoVector",
        )

    def delta_phi(self, other):
        """Compute difference in angle

        Returns a value within [-pi, pi)
        """
        return (self.phi - other.phi + numpy.pi) % (2 * numpy.pi) - numpy.pi


@awkward1.mixin_class(behavior)
class PolarTwoVector(TwoVector):
    """A polar coordinate 2-d vector

    This class overloads the properties r, phi with getitem accessors
    and provides properties x, y
    Some additional properties are overridden for performance
    """

    @property
    def x(self):
        return self.r * numpy.cos(self.phi)

    @property
    def y(self):
        return self.r * numpy.sin(self.phi)

    @property
    def r(self):
        return self["r"]

    @property
    def phi(self):
        return self["phi"]

    @property
    def r2(self):
        return self.r ** 2


@awkward1.mixin_class(behavior)
class ThreeVector(TwoVector):
    """A cartesian 3-dimensional vector

    A heavy emphasis towards a momentum vector interpretation is assumed.
    Properties this class requires: x, y, z
    """

    @property
    def pz(self):
        return self.z

    @property
    def rho2(self):
        return self.r2 + self.z ** 2

    @property
    def rho(self):
        return numpy.sqrt(self.rho2)

    @property
    def theta(self):
        return numpy.arctan2(self.r, self.z)

    @property
    def p2(self):
        return self.rho2

    @property
    def p(self):
        return self.rho

    @awkward1.mixin_class_method(numpy.absolute)
    def abs(self):
        return self.p

    @awkward1.mixin_class_method(numpy.add, {"ThreeVector"})
    def add(self, other):
        return awkward1.zip(
            {"x": self.x + other.x, "y": self.y + other.y, "z": self.z + other.z},
            with_name="ThreeVector",
        )

    def sum(self, axis=-1):
        out = awkward1.zip(
            {
                "x": awkward1.sum(self.x, axis=axis),
                "y": awkward1.sum(self.y, axis=axis),
                "z": awkward1.sum(self.z, axis=axis),
            },
            with_name="ThreeVector",
            highlevel=False,
        )
        return awkward1._util.wrap(out, cache=self.cache, behavior=self.behavior)

    @awkward1.mixin_class_method(numpy.multiply, {numbers.Number})
    def prod(self, other):
        return awkward1.zip(
            {"x": self.x * other, "y": self.y * other, "z": self.z * other},
            with_name="ThreeVector",
        )


@awkward1.mixin_class(behavior)
class SphericalThreeVector(ThreeVector, PolarTwoVector):
    """A spherical coordinate 3-d vector

    This class overloads the properties rho, theta with getitem accessors
    and provides properties r, z
    Some additional properties are overridden for performance
    """

    @property
    def r(self):
        return self.rho * numpy.sin(self.theta)

    @property
    def z(self):
        return self.rho * numpy.cos(self.theta)

    @property
    def rho(self):
        return self["rho"]

    @property
    def theta(self):
        return self["theta"]

    @property
    def p(self):
        return self.rho

    @property
    def p2(self):
        return self.rho ** 2


@awkward1.mixin_class(behavior)
class LorentzVector(ThreeVector):
    """A cartesian Lorentz vector

    A heavy emphasis towards a momentum vector interpretation is assumed.
    + - - - metric
    Properties this class requires: x, y, z, t
    """

    @property
    def energy(self):
        return self.t

    @property
    def eta(self):
        return numpy.arcsinh(self.z / self.r)

    @property
    def mass2(self):
        return self.t ** 2 - self.p2

    @property
    def mass(self):
        return numpy.sqrt(self.mass2)

    @awkward1.mixin_class_method(numpy.absolute)
    def abs(self):
        return self.mass

    @awkward1.mixin_class_method(numpy.add, {"LorentzVector"})
    def add(self, other):
        return awkward1.zip(
            {
                "x": self.x + other.x,
                "y": self.y + other.y,
                "z": self.z + other.z,
                "t": self.t + other.t,
            },
            with_name="LorentzVector",
        )

    def sum(self, axis=-1):
        out = awkward1.zip(
            {
                "x": awkward1.sum(self.x, axis=axis),
                "y": awkward1.sum(self.y, axis=axis),
                "z": awkward1.sum(self.z, axis=axis),
                "t": awkward1.sum(self.t, axis=axis),
            },
            with_name="LorentzVector",
            highlevel=False,
        )
        return awkward1._util.wrap(out, cache=self.cache, behavior=self.behavior)

    @awkward1.mixin_class_method(numpy.multiply, {numbers.Number})
    def prod(self, other):
        return awkward1.zip(
            {
                "x": self.x * other,
                "y": self.y * other,
                "z": self.z * other,
                "t": self.t * other,
            },
            with_name="LorentzVector",
        )

    def delta_r2(self, other):
        return (self.eta - other.eta) ** 2 + self.delta_phi(other) ** 2

    def delta_r(self, other):
        return numpy.sqrt(self.delta_r2(other))

    def nearest(
        self,
        other,
        axis=1,
        metric=lambda a, b: a.delta_r(b),
        return_metric=False,
        threshold=None,
    ):
        """Return nearest object to this one

        Finds item in ``other`` satisfying ``min(metric(self, other))``.
        The two arrays should be broadcast-compatible on all axes other than the specified
        axis, which will be used to form a cartesian product. If axis=None, broadcast arrays directly.
        The return shape will be that of ``self``.

        Parameters
        ----------
            other : awkward1.Array
                Another array with same shape in all but ``axis``
            axis : int, optional
                The axis to form the cartesian product (default 1). If None, the metric
                is directly evaluated on the input arrays (i.e. they should broadcast)
            metric : callable
                A function of two arguments, returning a scalar. The default metric is `delta_r`.
            return_metric : bool, optional
                If true, return both the closest object and its metric (default false)
            threshold : Number, optional
                If set, any objects with ``metric > threshold`` will be masked from the result
        """
        if axis is None:
            a, b = self, other
            # NotImplementedError: ak.firsts with axis=-1
            axis = other.layout.purelist_depth - 2
        else:
            a, b = awkward1.unzip(
                awkward1.cartesian([self, other], axis=axis, nested=True)
            )
        mval = metric(a, b)
        # prefer keepdims=True: awkward-1.0 #434
        mmin = awkward1.singletons(awkward1.argmin(mval, axis=axis + 1))
        out = awkward1.firsts(b[mmin], axis=axis + 1)
        metric = awkward1.firsts(mval[mmin], axis=axis + 1)
        if threshold is not None:
            out = out.mask[metric <= threshold]
        if return_metric:
            return out, metric
        return out


@awkward1.mixin_class(behavior)
class PtEtaPhiMLorentzVector(LorentzVector, SphericalThreeVector):
    """A Lorentz vector using pseudorapidity and mass

    This class overloads the pt, eta, phi, mass properties with getitem accessors
    and provides properties rho, theta, r, z, t
    Some additional properties are overridden for performance
    """

    @property
    def pt(self):
        return self["pt"]

    @property
    def eta(self):
        return self["eta"]

    @property
    def phi(self):
        return self["phi"]

    @property
    def mass(self):
        return self["mass"]

    @property
    def rho(self):
        return self.pt * numpy.cosh(self.eta)

    @property
    def theta(self):
        raise NotImplementedError

    @property
    def r(self):
        return self.pt

    @property
    def z(self):
        return self.pt * numpy.sinh(self.eta)

    @property
    def t(self):
        return numpy.hypot(self.rho, self.mass)

    @property
    def rho2(self):
        return self.rho ** 2

    @property
    def mass2(self):
        return self.mass ** 2

    @awkward1.mixin_class_method(numpy.multiply, {numbers.Number})
    def prod(self, other):
        return awkward1.zip(
            {
                "pt": self.pt * other,
                "eta": self.eta,
                "phi": self.phi,
                "mass": self.mass * other,
            },
            with_name="PtEtaPhiMLorentzVector",
        )


@awkward1.mixin_class(behavior)
class PtEtaPhiELorentzVector(LorentzVector, SphericalThreeVector):
    """A Lorentz vector using pseudorapidity and energy

    This class overloads the pt, eta, phi, energy, t properties with getitem accessors
    and provides properties rho, theta, r, z
    Some additional properties are overridden for performance
    """

    @property
    def pt(self):
        return self["pt"]

    @property
    def eta(self):
        return self["eta"]

    @property
    def phi(self):
        return self["phi"]

    @property
    def energy(self):
        return self["energy"]

    @property
    def t(self):
        return self["energy"]

    @property
    def rho(self):
        return self.pt * numpy.cosh(self.eta)

    @property
    def theta(self):
        raise NotImplementedError

    @property
    def r(self):
        return self.pt

    @property
    def z(self):
        return self.pt * numpy.sinh(self.eta)

    @property
    def rho2(self):
        return self.rho ** 2

    @awkward1.mixin_class_method(numpy.multiply, {numbers.Number})
    def prod(self, other):
        return awkward1.zip(
            {
                "pt": self.pt * other,
                "eta": self.eta,
                "phi": self.phi,
                "energy": self.energy * other,
            },
            with_name="PtEtaPhiELorentzVector",
        )



import warnings
import weakref
import json
import awkward1
import uproot4 as uproot
from coffea.nanoevents.util import quote, key_to_tuple, tuple_to_key
from coffea.nanoevents.mapping import TrivialOpener, UprootSourceMapping, CachedMapping
from coffea.nanoevents.schemas import BaseSchema, NanoAODSchema


class NanoEventsFactory:
    """A factory class to build NanoEvents objects"""

    def __init__(self, schema, mapping, partition_key, cache=None):
        self._schema = schema
        self._mapping = mapping
        self._partition_key = partition_key
        self._cache = cache
        self._events = lambda: None

    def __getstate__(self):
        return {
            "schema": self._schema,
            "mapping": self._mapping,
            "partition_key": self._partition_key,
        }

    def __setstate__(self, state):
        self._schema = state["schema"]
        self._mapping = state["mapping"]
        self._partition_key = state["partition_key"]
        self._cache = None
        self._events = lambda: None

    @classmethod
    def from_file(
        cls,
        file,
        treepath="/Events",
        entry_start=None,
        entry_stop=None,
        runtime_cache=None,
        persistent_cache=None,
        schemaclass=NanoAODSchema,
        metadata=None,
        uproot_options={},
        access_log=None,
    ):
        """Quickly build NanoEvents from a file

        Parameters
        ----------
            file : str or uproot.reading.ReadOnlyDirectory
                The filename or already opened file using e.g. ``uproot.open()``
            treepath : str, optional
                Name of the tree to read in the file
            entry_start : int, optional
                Start at this entry offset in the tree (default 0)
            entry_stop : int, optional
                Stop at this entry offset in the tree (default end of tree)
            runtime_cache : dict, optional
                A dict-like interface to a cache object. This cache is expected to last the
                duration of the program only, and will be used to hold references to materialized
                awkward1 arrays, etc.
            persistent_cache : dict, optional
                A dict-like interface to a cache object. Only bare numpy arrays will be placed in this cache,
                using globally-unique keys.
            schemaclass : BaseSchema
                A schema class deriving from `BaseSchema` and implementing the desired view of the file
            metadata : dict, optional
                Arbitrary metadata to add to the `base.NanoEvents` object
            uproot_options : dict, optional
                Any options to pass to ``uproot.open``
            access_log : list, optional
                Pass a list instance to record which branches were lazily accessed by this instance
        """
        if not issubclass(schemaclass, BaseSchema):
            raise RuntimeError("Invalid schema type")
        if isinstance(file, str):
            tree = uproot.open(file, **uproot_options)[treepath]
        elif isinstance(file, uproot.reading.ReadOnlyDirectory):
            tree = file[treepath]
        elif "<class 'uproot.rootio.ROOTDirectory'>" == str(type(file)):
            raise RuntimeError(
                "The file instance (%r) is an uproot3 type, but this module is only compatible with uproot4 or higher"
                % file
            )
        else:
            raise TypeError("Invalid file type (%s)" % (str(type(file))))
        if entry_start is None or entry_start < 0:
            entry_start = 0
        if entry_stop is None or entry_stop > tree.num_entries:
            entry_stop = tree.num_entries
        partition_tuple = (
            str(tree.file.uuid),
            tree.object_path,
            "{0}-{1}".format(entry_start, entry_stop),
        )
        uuidpfn = {partition_tuple[0]: tree.file.file_path}
        mapping = UprootSourceMapping(
            TrivialOpener(uuidpfn, uproot_options), access_log=access_log
        )
        mapping.preload_tree(partition_tuple[0], partition_tuple[1], tree)
        if persistent_cache is not None:
            mapping = CachedMapping(persistent_cache, mapping)
        base_form = cls._extract_base_form(tree)
        if metadata is not None:
            base_form["parameters"]["metadata"] = metadata
        schema = schemaclass(base_form)
        return cls(schema, mapping, tuple_to_key(partition_tuple), cache=runtime_cache)

    def __len__(self):
        uuid, treepath, entryrange = key_to_tuple(self._partition_key)
        start, stop = (int(x) for x in entryrange.split("-"))
        return stop - start

    @classmethod
    def _extract_base_form(cls, tree):
        branch_forms = {}
        for key, branch in tree.iteritems():
            if "," in key or "!" in key:
                warnings.warn(
                    f"Skipping {key} because it contains characters that NanoEvents cannot accept [,!]"
                )
                continue
            if len(branch):
                continue
            form = branch.interpretation.awkward_form(None)
            form = uproot._util.awkward_form_remove_uproot(awkward1, form)
            form = json.loads(form.tojson())
            if (
                form["class"].startswith("ListOffset")
                and form["content"]["class"] == "NumpyArray"  # noqa
            ):
                form["form_key"] = quote(f"{key},!load")
                form["content"]["form_key"] = quote(f"{key},!load,!content")
                form["content"]["parameters"] = {"__doc__": branch.title}
            elif form["class"] == "NumpyArray":
                form["form_key"] = quote(f"{key},!load")
                form["parameters"] = {"__doc__": branch.title}
            else:
                warnings.warn(
                    f"Skipping {key} as it is not interpretable by NanoEvents"
                )
                continue
            branch_forms[key] = form

        return {
            "class": "RecordArray",
            "contents": branch_forms,
            "parameters": {"__doc__": tree.title},
            "form_key": "",
        }

    def events(self):
        """Build events"""
        events = self._events()
        if events is None:
            behavior = dict(self._schema.behavior)
            behavior["__events_factory__"] = self
            events = awkward1.from_arrayset(
                self._schema.form,
                self._mapping,
                prefix=self._partition_key,
                sep="/",
                lazy=True,
                lazy_lengths=len(self),
                lazy_cache="attach" if self._cache is None else self._cache,
                behavior=behavior,
            )
            self._events = weakref.ref(events)

        return events



from .NanoTestProcessor import NanoTestProcessor
from .NanoEvents0Processor import NanoEvents0Processor
from .NanoEventsProcessor import NanoEventsProcessor

__all__ = ["NanoTestProcessor", "NanoEvents0Processor", "NanoEventsProcessor"]



"""Awkward 0.x NanoAOD collection methods

This package provides the object method mixins for the various
types of NanoAOD collections, as well as a default mapping for collection
names to methods.
"""
from .common import METVector, LorentzVector, Candidate
from .leptons import Electron, Muon, Photon, FsrPhoton, Tau
from .jets import Jet, FatJet
from .generator import GenParticle, GenVisTau


collection_methods = {
    'CaloMET': METVector,
    'ChsMET': METVector,
    'GenMET': METVector,
    'MET': METVector,
    'METFixEE2017': METVector,
    'PuppiMET': METVector,
    'RawMET': METVector,
    'TkMET': METVector,
    # pseudo-lorentz: pt, eta, phi, mass=0
    'IsoTrack': LorentzVector,
    'SoftActivityJet': LorentzVector,
    'TrigObj': LorentzVector,
    # True lorentz: pt, eta, phi, mass
    'FatJet': FatJet,
    'GenDressedLepton': LorentzVector,
    'GenJet': LorentzVector,
    'GenJetAK8': LorentzVector,
    'Jet': Jet,
    'LHEPart': LorentzVector,
    'SV': LorentzVector,
    'SubGenJetAK8': LorentzVector,
    'SubJet': LorentzVector,
    'PFCands': LorentzVector,  # available in NanoAODJMAR
    # Candidate: LorentzVector + charge
    'Electron': Electron,
    'Muon': Muon,
    'Photon': Photon,
    'Tau': Tau,
    'GenVisTau': GenVisTau,
    'FsrPhoton': FsrPhoton,
    # special
    'GenPart': GenParticle,
}

__all__ = [
    'METVector',
    'LorentzVector',
    'Candidate',
    'Electron',
    'Muon',
    'Photon',
    'FsrPhoton',
    'Tau',
    'Jet',
    'FatJet',
    'GenParticle',
    'GenVisTau',
    'collection_methods',
]



from collections.abc import MutableMapping
import awkward1
import uproot4


class LazyDataFrame(MutableMapping):
    """Simple delayed uproot reader (a la lazyarrays)

    One can access branches either through ``df["bname"]`` or ``df.bname``, although
    the latter is restricted to branches that do not start with a leading underscore.
    Keeps track of values accessed, in the `materialized` attribute.

    Parameters
    ----------
        tree : uproot4.TTree
            Tree to read
        entrystart : int, optional
            First entry to read, default: 0
        entrystop : int, optional
            Last entry to read, default None (read to end)
        preload_items : iterable
            Force preloading of a set of columns from the tree
        flatten : bool
            Remove jagged structure from columns read
    """

    def __init__(
        self, tree, entrystart=None, entrystop=None, preload_items=None, flatten=False
    ):
        self._tree = tree
        self._flatten = flatten
        self._branchargs = {
            "decompression_executor": uproot4.source.futures.TrivialExecutor(),
            "interpretation_executor": uproot4.source.futures.TrivialExecutor(),
        }
        if entrystart is None or entrystart < 0:
            entrystart = 0
        if entrystop is None or entrystop > tree.num_entries:
            entrystop = tree.num_entries
        self._branchargs["entry_start"] = entrystart
        self._branchargs["entry_stop"] = entrystop
        self._available = {k for k in self._tree.keys()}
        self._dict = {}
        self._materialized = set()
        if preload_items:
            self.preload(preload_items)

    def __delitem__(self, key):
        del self._dict[key]

    def __getitem__(self, key):
        if key in self._dict:
            return self._dict[key]
        elif key in self._tree:
            self._materialized.add(key)
            array = self._tree[key].array(**self._branchargs)
            if self._flatten and isinstance(awkward1.type(array).type, awkward1.types.ListType):
                array = awkward1.flatten(array)
            array = awkward1.to_awkward0(array)
            self._dict[key] = array
            return self._dict[key]
        else:
            raise KeyError(key)

    def __getattr__(self, key):
        if key.startswith("_"):
            raise AttributeError(key)
        try:
            return self.__getitem__(key)
        except KeyError:
            raise AttributeError(key)

    def __iter__(self):
        for item in self._available:
            yield item

    def __len__(self):
        return len(self._dict)

    def __setitem__(self, key, value):
        self._dict[key] = value

    def __contains__(self, key):
        # by default, MutableMapping uses __getitem__ to test, but we want to avoid materialization
        return key in self._dict or key in self._tree

    @property
    def available(self):
        """Set of available columns"""
        return self._available

    @property
    def columns(self):
        """Set of available columns"""
        return self._available

    @property
    def materialized(self):
        """Set of columns read from tree"""
        return self._materialized

    @property
    def size(self):
        """Length of column vector"""
        return self._branchargs["entry_stop"] - self._branchargs["entry_start"]

    def preload(self, columns):
        """Force loading of several columns

        Parameters
        ----------
            columns : iterable
                A list of columns to load
        """
        for name in columns:
            if name in self._tree:
                _ = self[name]


class PreloadedDataFrame(MutableMapping):
    """A dataframe for instances like spark where the columns are preloaded

    Provides a unified interface, matching that of LazyDataFrame.

    Parameters
    ----------
        size : int
            Number of rows
        items : dict
            Mapping of column name to column array
    """

    def __init__(self, size, items):
        self._size = size
        self._dict = items
        self._accessed = set()

    def __delitem__(self, key):
        del self._dict[key]

    def __getitem__(self, key):
        self._accessed.add(key)
        return self._dict[key]

    def __getattr__(self, key):
        if key.startswith("_"):
            raise AttributeError(key)
        try:
            return self.__getitem__(key)
        except KeyError:
            raise AttributeError(key)

    def __iter__(self):
        for key in self._dict:
            yield key

    def __len__(self):
        return len(self._dict)

    def __setitem__(self, key, value):
        self._dict[key] = value

    @property
    def available(self):
        """List of available columns"""
        return self._dict.keys()

    @property
    def materialized(self):
        """List of accessed columns"""
        return self._accessed

    @property
    def size(self):
        """Length of column vector"""
        return self._size
